/*
 *   Copyright 2011 Institute of Communication and Computer Systems (ICCS) 
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

#include <dmmlib/dmmlib.h>
#include "dmm_config.h"
#include "other.h"
#ifdef HAVE_LOCKS
#include "posix_lock.h"
#endif /* HAVE_LOCKS */
#if defined (COALESCING_FIXED) || defined (COALESCING_VARIABLE)
#include "coalesce.h"
#endif /* COALESCING_FIXED || COALESCING_VARIABLE */
#include "dmm_adaptor.h"
#include "block_header.h"

void custom_ahfree(allocator_t *allocator, heap_t* heap, void *ptr) {
    size_t size;
    int fixed_list_id, i;
    maptable_node_t *current_maptable_node;
#if defined (COALESCING_FIXED) || defined (COALESCING_VARIABLE)
    bool coalesced;
    size_t max_coal_size;
#endif /* COALESCING_FIXED || COALESCING_VARIABLE */

#ifndef WITH_MEMORY_SPACE_AWARENESS
    int heap_id;
    /* Currently all the memory space aware allocators are pre-initialized */
    if(allocator == NULL) {
        allocator = &systemallocator;
    }

    if(heap == NULL) {
        heap_id = map_thread_heap();
        heap = &allocator->heaps[heap_id];
    }
#endif /* WITH_MEMORY_SPACE_AWARENESS */

    size = get_size(ptr);

#ifdef HAVE_LOCKS	
    posix_lock(heap);
#endif /* HAVE_LOCKS */

    remove_block(ptr, heap->used_blocks_head);

#if defined (COALESCING_FIXED) || defined (COALESCING_VARIABLE)
    if(is_previous_free(ptr)) {
        if(get_owner(ptr) == get_owner(get_dlprevious(ptr))) {

#ifdef COALESCING_FIXED
            max_coal_size = MAX_COALESCE_SIZE;
#endif /* COALESCING_FIXED */
#ifdef COALESCING_VARIABLE
            max_coal_size = heap->dmm_knobs.max_coalesce_size;
#endif /* COALESCING_VARIABLE */

            if(get_previous_size(ptr) + get_size(ptr) + HEADER_SIZE <=
                    max_coal_size) {
                ptr = coalesce(ptr, heap, allocator);
                coalesced = true;
            } else {
                mark_free(ptr);
                coalesced = false;
            }
        }
    } else {
        mark_free(ptr);
        coalesced = false;
    }
#endif /* COALESCING_FIXED || COALESCING_VARIABLE */

    /* Check if the block could be put in a fixed list */
    fixed_list_id = map_size_to_list(heap, size);

    if(fixed_list_id != -1) {
        current_maptable_node = heap->maptable_head;		
        if(fixed_list_id != 0) {
            for(i = 1; i < fixed_list_id; i++) {
                current_maptable_node = current_maptable_node->next;
            }
        }
#ifdef BLOCKS_IN_DLL
        set_previous(ptr, NULL);
        set_previous(current_maptable_node->fixed_list_head, ptr);
#endif /* BLOCKS_IN_DLL */
        set_next(ptr, current_maptable_node->fixed_list_head);
        current_maptable_node->fixed_list_head = ptr;
    } else { /* put it in the free list */
#if defined (COALESCING_FIXED) || defined (COALESCING_VARIABLE)
        /* If the block is coalesced, there is no need to add it to the free
         * list as there is already an entry of the old block
         */
        if(!coalesced) {
#ifdef BLOCKS_IN_DLL
            set_previous(heap->free_list_head, ptr);
            set_previous(ptr, NULL);
#endif /* BLOCKS_IN_DLL */
            set_next(ptr, heap->free_list_head);
            heap->free_list_head = ptr;
        }
#else
#ifdef BLOCKS_IN_DLL
        set_previous(heap->free_list_head, ptr);
        set_previous(ptr, NULL);
#endif /* BLOCKS_IN_DLL */
        set_next(ptr, heap->free_list_head);
        heap->free_list_head = ptr;
#endif /* COALESCING_FIXED || COALESCING_VARIABLE */
    }

    /* Begin of Stats */

    heap->dmm_stats.live_objects -= 1;
    heap->dmm_stats.num_free += 1;

    /* End of Stats */

#ifdef WITH_KNOBS
    /* Refresh the state of the heap allocator if a certain number of
     * free's has been served already
     */
    /* TODO Define 100 as a constant */
    if(heap->dmm_stats.num_free % 100) {
        free_state_refresh(heap);
    }
#endif /* WITH_KNOBS */

#ifdef HAVE_LOCKS
    posix_unlock(heap);
#endif /* HAVE_LOCKS */
}

/* Currently all the memory space aware allocators are pre-initialized, so
 * we do not expect any custom_ahmalloc call without knowing which allocator
 * and heap are to be used.
 */
#ifndef WITH_MEMORY_SPACE_AWARENESS
void custom_free(void *ptr) {
    custom_ahfree(NULL, NULL, ptr);
}
#endif /* WITH_MEMORY_SPACE_AWARENESS */

